/*
 * Copyright 2016 Michal Žejdl
 * input: table mibfdbport
 * output: network topology (table peer, graphviz...)
 *
 * principles
 * 	- continuous input from the oldest entry up to the present (now - 1 sec)
 * 		(every record can help to reveal current topology)
 * 	- it is from mysql table now, but can be modified to read from stdin
 * 	- the algorithm is created for one atomic entry
 * 		(bridge - port - MAC address)
 * 	- data are handled as strings (except the time)
 * 	- the primary data structure is singly linked list
 * 	- every bridge and MAC address are stored as sole instance
 */

#include "../lib/log.h"
#include "../lib/db.h"

#include "input.h"
#include "process.h"

// printf()
#include <stdio.h>

// basename()
#include <libgen.h>

// sleep()
#include <unistd.h>

// malloc(), free()
#include <stdlib.h>

// str*()
#include <string.h>

// time()
#include <time.h>

// limit amount of returned rows
#define ROW_LIMIT	10000

// aging interval in seconds (similar to switch MAC address aging time)
// after this time the old data are removed from the graph
#define AGING_TIME	3600

void print_usage(char *path)
{
	char *name = basename(path);
	fprintf(stderr, "%s read FDB from database and make a network topology graph\n", name);
	fprintf(stderr, "Usage: %s <host> <user> <password> <db>\n", name);
}

// fill string with spaces
void wipe_string(char *string)
{
	unsigned i;
	
	for (i = 0; i < strlen(string); i++)
		string[i] = ' ';
}

int main (int argc, char **argv)
{
	//int connected = 0;
	time_t time_from, time_till, now, range = 0;

	MYSQL_RES *data;

	// init global variable
	logfile = stderr;

	if (argc != 5) {
		print_usage(argv[0]);
		return 1;
	}

	// pass args to db unit
	if (db_init(argv[1], argv[2], argv[3], argv[4])) {
		log_write("fatal: db_init() failed\n");
		return 1;
	}

	// hide password from system
	wipe_string(argv[3]);

	// when in the past start with the data
	//time_from = get_start_time();
	//time_from = time(NULL);
	// the most reasonable start time
	time_from = time(NULL) - AGING_TIME;

	printf("start time: %ld\n", time_from);

	// profiling
	/*
	int i = 1000;
	while (i--) {
	*/

	while (1) {

		/*
		while (! connected) {
			log_write("still not connected to db, trying...\n");
			if ((connected = db_connect())) {
				log_write("connected, go on\n");
				break;
			}
			sleep(10);
		}
		*/

		now = time(NULL);

		printf("now %ld from %ld ", now, time_from);

		// shrink range to fit below the now
		while (time_from + range >= now && range > 0)
			range >>= 1;

		// set upper endpoint
		time_till = time_from + range;

		printf("range %ld till %ld ", range, time_till);

		// get close to actual time to get most recent records
		if (time_till < now) {

			// no time to sleep, hurry for the data
			data = get_data(time_from, time_till);

			if (data) {

				printf("rows %lld", mysql_num_rows(data));

				if (mysql_num_rows(data) < ROW_LIMIT)
					/*
					 * adjust range
					 * to get more data next time
					 */
					range = range ? range << 1 : 1;
				else
					/*
					 * shrink range to get less data
					 * next time
					 */
					range >>= 1;

				process_data(data);

				// dispose data
				mysql_free_result(data);

			} else {

				puts("\n");

				/*
				 * NULL pointer means an error
				 * shrink the range to avoid possible malloc()
				 * problems in myslq_store_result()
				 */
				range >>= 1;
				/*
				 * and wait a while - there may be connection
				 * errors whose time can heal
				 */
				sleep(10);
			}

			// move to the next interval
			time_from = time_till + 1;

			// remove aged data
			aging(time_from - AGING_TIME);

		} else {

			puts("\n");

			process_topology();

			// do the discovery only once
			//break;

			// wait long enough to see this delay in collectd
			//sleep(1);
			sleep(300);
		}
				
		printf("range %ld ", range);
	}

	db_finish();

	return 0;
}
